// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.applet.*;
import java.awt.*;

public class LatheApplet extends Applet implements Runnable
{
	/** Model waiting to move state. */
	private final int STATE_WAIT = 0;
	/** Model moving away state. */
	private final int STATE_AWAY = 1;
	/** Model moving toward state. */
	private final int STATE_TOWARDS = 2;
	
	/** Number of frames model will wait. */
	private final int WAIT_LENGTH = 300;
	/** Number of frames model will move. */
	private final int MOVE_LENGTH = 250;

	/** Distance model will move. */
	private final float MOVEMENT_DISTANCE = 1200;
	/** Distance from origin model will wait. */
	private final float REST_DISTANCE = 20;

	/** Scale of cube model. */
	private final int CUBE_SIZE = 9;
	/** Scale of pyramid model. */
	private final int PYRAMID_SIZE = 10;
	
	/** Number of times to rotate top lathe vertices. */
	private final int NUM_TOP_ROTATIONS = 90;
	/** Bounding spinning-top size of model. */
	private final int TOP_SIZE = 12;	
	
	/** Number of times to rotate cylinder lathe vertices. */
	private final int NUM_CYLINDER_ROTATIONS = 90;
	/** Scale of cylinder model. */
	private final int CYLINDER_SIZE = 10;

	/** Number of times to rotate cone lathe vertices. */
	private final int NUM_CONE_ROTATIONS = 90;
	/** Bounding cone size of models. */
	private final int CONE_SIZE = 11;
	
	/** Number of times to rotate sphere lathe vertices. */
	private final int NUM_SPHERE_ROTATIONS = 20;
	/** Number of points on sphere lathe. */
	private final int NUM_SPHERE_LATHE_VERTICES = 20;
	/** Scale of sphere model. */
	private final int SPHERE_SIZE = 12;
	
	/** Frequency model will rotate on its x-axis. */
	private final float X_ROTATE_FREQUENCY = +0.010f;
	/** Frequency model will rotate on its y-axis. */
	private final float Y_ROTATE_FREQUENCY = -0.008f;
	/** Frequency model will rotate on its z-axis. */
	private final float Z_ROTATE_FREQUENCY = +0.004f;

	/** Direction light-source is shining. */
	private final Vector3d LIGHT_DIRECTION = new Vector3d(0.3f, 0.3f, -1f).normalise();
	/** Amount of glint light-source provides. */
	private final float LIGHT_GLINT = 0.95f;
	/** Field of Vision angle size. */
	private final float FOV = (float)(50 * (Math.PI / 180));

	/** Filename of background image. */
	private final String BACKGROUND_IMAGE_FILENAME = "background.jpg";

	/** Thread used for updating application. */
	private Thread updateThread;
	
	/** Double buffer. */
	private Image offscreen;
	/** Background image. */
	private Image background;
	
	/** List of models. */
	private Model3d[] modelList;
	
	/** Current model being shown. */
	private Model3d model;
	
	/** Whether background image has loaded. */
	private boolean backgroundLoaded = false;

	/** Current model being viewed. */
	private int currentModel = 0;
	/** Current model state. */
	private int state = STATE_TOWARDS;
	/** Current model distance from rest distance. */
	private float moveDelta = MOVEMENT_DISTANCE;
	/** Number of frames waited in wait state. */
	private int framesWaited = 0;
	/** Number of frames in moving towards/away state. */
	private int framesMoving = 0;
	/** Current frame. */
	private int frame = 0;

	public void init()
	{
		// Create double buffer
		offscreen  = createImage(getSize().width, getSize().height);

		// Common constant
		final float HALF_ROOT_TWO = (float)Math.sqrt(2) / 2;

		// Create cube model
		Vector3d[]  cubeLatheVertexList = {new Vector3d(HALF_ROOT_TWO, +0.5f, 0), new Vector3d(HALF_ROOT_TWO, -0.5f, 0)};
		Transform3d cubeTransform       =  new Transform3d().scale(new Vector3d(CUBE_SIZE, CUBE_SIZE, CUBE_SIZE));
		Model3d     cube                =  new Model3d(4, cubeLatheVertexList, Color.red).transform(cubeTransform);

		// Create pyramid model
		Vector3d[]  pyramidLatheVertexList = {new Vector3d(HALF_ROOT_TWO, +0.5f, 0)};
		Transform3d pyramidTransform       =  new Transform3d().translate(new Vector3d(0, -0.2f, 0)).scale(new Vector3d(PYRAMID_SIZE, PYRAMID_SIZE, PYRAMID_SIZE));
		Model3d     pyramid                =  new Model3d(4, pyramidLatheVertexList, Color.blue).transform(pyramidTransform);

		// Create cylinder model
		Vector3d[]  cylinderLatheVertexList = {new Vector3d(+0.5f, +0.5f, 0), new Vector3d(+0.5f, -0.5f, 0)};
		Transform3d cylinderTransform       =  new Transform3d().scale(new Vector3d(CYLINDER_SIZE, CYLINDER_SIZE, CYLINDER_SIZE));
		Model3d     cylinder                =  new Model3d(NUM_CYLINDER_ROTATIONS, cylinderLatheVertexList, Color.green).transform(cylinderTransform);

		// Create cone model
		Vector3d[]  coneLatheVertexList = {new Vector3d(+0.5f, +0.5f, 0)};
		Transform3d coneTransform      =  new Transform3d().translate(new Vector3d(0, -0.2f, 0)).scale(new Vector3d(CONE_SIZE, CONE_SIZE, CONE_SIZE));
		Model3d     cone                =  new Model3d(NUM_CONE_ROTATIONS, coneLatheVertexList, Color.yellow).transform(coneTransform);

		// Create spinning-top model
		Vector3d[]  topLatheVertexList = {new Vector3d(+0.5f, 0, 0)};
		Transform3d topTransform       =  new Transform3d().scale(new Vector3d(TOP_SIZE, TOP_SIZE, TOP_SIZE));
		Model3d     top                =  new Model3d(NUM_TOP_ROTATIONS, topLatheVertexList, Color.cyan).transform(topTransform);

		// Sphere vertex lathe list
		Vector3d[] sphereLatheVertexList = new Vector3d[NUM_SPHERE_LATHE_VERTICES];

		// For each lathe vertex
		for (int i = 0; i < sphereLatheVertexList.length; ++i)
		{
			// Calculate current stage of circle
			float stage = (float)(i + 1) / (sphereLatheVertexList.length + 1);
			// Calculate currnet lathe vertex of circle
			sphereLatheVertexList[i] = new Vector3d((float)Math.sin(stage * Math.PI) * 0.5f, (float)Math.cos(stage * Math.PI) * 0.5f , 0);
		}

		// Create circle model
		Transform3d sphereTransform = new Transform3d().scale(new Vector3d(SPHERE_SIZE, SPHERE_SIZE, SPHERE_SIZE));
		Model3d     sphere          = new Model3d(NUM_SPHERE_ROTATIONS, sphereLatheVertexList, Color.magenta).transform(sphereTransform);

		// Set model list to be viewed
		Model3d[] list = {cube, pyramid, cylinder, cone, sphere, top};
		modelList = list;
	}

	public void start()
	{
		// Create update thread
		updateThread = new Thread(this);
		// Start update thread
		updateThread.start();
	}

	public void stop()
	{
		// Stop update thread
		updateThread = null;
	}

	public void destroy()
	{
		// Stop update thread
		updateThread = null;
	}

	public void loadBackgroundImage()
	{
		// Set to border layout
		setLayout(new BorderLayout());
		// Store wait message label
		Label messageLabel = new Label("Loading image...", Label.CENTER);
		// Add label to center of applet
		add(messageLabel, BorderLayout.CENTER);
		// Show message
		repaint();

		// Background image ID for media tracker
		final int BACKGROUND_IMAGE_ID = 0;
		// Get background image
		background = getImage(getDocumentBase(), BACKGROUND_IMAGE_FILENAME);					

		// Media tracker for wait for image loading
		MediaTracker mediaTracker = new MediaTracker(this);
		// Add image to media tracker
		mediaTracker.addImage(background, BACKGROUND_IMAGE_ID);
		
		try
		{
			// Wait for background image to load
			mediaTracker.waitForID(BACKGROUND_IMAGE_ID);
		}
		catch (InterruptedException e)
		{
		}

		// If error loading background image
		if (mediaTracker.isErrorID(BACKGROUND_IMAGE_ID))
		{
			throw new RuntimeException("Error trying to load " + BACKGROUND_IMAGE_FILENAME);
		}

		// Remove loading message
		remove(messageLabel);

		// Background has been loaded
		backgroundLoaded = true;
	}

	public void updateModelMovement()
	{
		// Check model movement state
		switch (state)
		{
			case STATE_WAIT:
			{
				// Waited for another frame
				++framesWaited;

				// If waited for long enough
				if (framesWaited >= WAIT_LENGTH)
				{
					// Reset waited counter
					framesWaited = 0;
					// Change to moving away state
					state = STATE_AWAY;
				}
				break;
			}

			case STATE_AWAY:
			{
				// Moved for another frame
				++framesMoving;
				
				// Calculate radians on current position sine wave
				float radians = ((float)framesMoving / MOVE_LENGTH) * ((float)Math.PI / 2);
				// Calculate current position
				moveDelta = MOVEMENT_DISTANCE - (float)(Math.cos(radians) * MOVEMENT_DISTANCE);

				// If moved far enough away
				if (framesMoving >= MOVE_LENGTH)
				{
					// Reset moving counter
					framesMoving = 0;
					// Cycle to next model
					currentModel = (currentModel + 1) % modelList.length;
					// Change to moving toward state
					state = STATE_TOWARDS;
				}

				break;
			}

			case STATE_TOWARDS:
			{
				// Moved for another frame
				++framesMoving;
				
				// Calculate radians on current position sine wave
				float radians = ((float)framesMoving / MOVE_LENGTH) * ((float)Math.PI / 2);
				// Calculate current position
				moveDelta = MOVEMENT_DISTANCE - (float)Math.sin(radians) * MOVEMENT_DISTANCE;

				// If moved close enough
				if (framesMoving >= MOVE_LENGTH)
				{
					// Reset moving counter
					framesMoving = 0;
					// Chnge to wait state
					state = STATE_WAIT;						
				}

				break;
			}
		}
	}
	
	public void run()
	{
		// Load background image
		loadBackgroundImage();

		// While executing thread is this one
		while (Thread.currentThread() == updateThread)
		{
			// Update model movement
			updateModelMovement();
			
			// Make copy of original model positioned on origin
			model = new Model3d(modelList[currentModel]);

			// Calculate model rotations for this frame
			float xRotate = (float)(Math.sin(frame * X_ROTATE_FREQUENCY) * 2 * Math.PI);
			float yRotate = (float)(Math.sin(frame * Y_ROTATE_FREQUENCY) * 2 * Math.PI);
			float zRotate = (float)(Math.sin(frame * Z_ROTATE_FREQUENCY) * 2 * Math.PI);
			
			// Calculate translation vector
			Vector3d translate = new Vector3d(0, 0, state != STATE_WAIT ? REST_DISTANCE + moveDelta : REST_DISTANCE);
			// Create model transformation
			Transform3d transform = new Transform3d().rotate(new Vector3d(xRotate, yRotate, zRotate)).translate(translate);
			// Transform model
			model.transform(transform);
			
			// Frame finished 
			++frame;

			// Update screen
			repaint();

			try
			{
				Thread.sleep(10);
			}
			catch (InterruptedException e)
			{
			}
		}
	}

	public void update(Graphics g)
	{
		// If background has loaded
		if (backgroundLoaded)
		{
			// Paint double buffer
			paint(g);
		}
		else
		{
			// Paint any components
			paintAll(g);
		}
	}

	public void paint(Graphics g)
	{
		// If background has loaded
		if (backgroundLoaded)
		{
			// Draw background
			offscreen.getGraphics().drawImage(background, 0, 0, null);
			// Draw model
			model.draw(offscreen.getGraphics(), getSize(), LIGHT_DIRECTION, LIGHT_GLINT, FOV);

			// Paint double buffer
			g.drawImage(offscreen, 0, 0, null);
		}

	}

	public String getAppletInfo()
	{
		return "Lathe - Copyright (c) 2001 Sean Wilson. All Rights Reserved.";
	}
}