// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.*;

/** Represents a 3D model composed of vertices and colored triangles. */
public class Model3d
{
	/** List of triangles making up model */
	private Triangle3d[] triangleList;
	/** List of vertices triangles are made from */
	private Vector3d[] vertexList;

	/** Represents a colored triangle composed of indices to model vertices. Vertices should be in clockwise order to aid back-face culling. */	 
	class Triangle3d
	{
		/** Triangle color. */
		public Color color;
		/** Model vertex index 1. */
		public int vertex1;
		/** Model vertex index 2. */
		public int vertex2;
		/** Model vertex index 3. */
		public int vertex3;

		/** Contructs a triangle.
		 *  @param initColor   triangle color.
		 *  @param initVertex1 model vertex index 1.
		 *  @param initVertex2 model vertex index 2.
		 *  @param initVertex3 model vertex index 3.
		 */
		public Triangle3d(Color initColor, int initVertex1, int initVertex2, int initVertex3)
		{
			color   = new Color(initColor.getRGB());
			vertex1 = initVertex1;
			vertex2 = initVertex2;
			vertex3 = initVertex3;
		}
	}

	/** Constructs a model from another.
	 *  @param  model3d model to clone.
	 */
	public Model3d(Model3d model3d)
	{
		// Copy triangle and vertex lists
		triangleList = (Triangle3d[])model3d.triangleList.clone();
		vertexList   = (  Vector3d[])model3d.vertexList.clone();
	}

	/** Uses a lathe technique to create a model. Model begins with a vertices at (0, +0.5, 0) and (0, -0.5, 0). Vertices provided are rotated around the y-axis and added
	 *  the specified number of times. Triangles are then added to created an enclosed model.
	 *  @param  numRotations    number of times to rotate provided vertices.
	 *  @param  latheVertexList list of vertices to be rotated.
	 *  @param  color           model color.
	 *  @return                 created model.
	 */
	public Model3d(int numRotations, Vector3d[] latheVertexList, Color color)
	{
		// Model vertex list
		vertexList = new Vector3d[numRotations * latheVertexList.length + 2];

		// Index of top model vertex
		final int TOP = 0;
		// Index of bottom model vertex
		final int BOTTOM = 1;
		
		// Add top and bottom model vertices
		vertexList[TOP   ] = new Vector3d( 0, +0.5f, 0);
		vertexList[BOTTOM] = new Vector3d( 0, -0.5f, 0);

		// Current vertex being added
		int currentVertex = 2;

		// For each rotation
		for (int i = 0; i < numRotations; ++i)
		{
			// For each lathe vertex
			for (int p = 0; p < latheVertexList.length; ++p)
			{
				// Calculate current stage of rotation
				float stage = (float)i / numRotations;
				// Add rotated vertex
				vertexList[currentVertex++] = latheVertexList[p].rotate(new Vector3d(0, stage * 2 * (float)Math.PI, 0));
			}
		}

		// Model triangle list
		triangleList = new Triangle3d[numRotations * 2 + numRotations * 2 * (latheVertexList.length - 1)];

		// Current triangle being added
		int currentTriangle = 0;

		// For each rotation
		for (int i = 0; i < numRotations; ++i)
		{
			// Diagram of top triangle being added:
			//
			//      TOP
			//     /|\
			//    /*| \
			//   /**|  \
			//  /---|---\
			//  a   c
			
			// Calculate vertex indices
			int a = ((i + 0) % numRotations) * latheVertexList.length + 2;
			int c = ((i + 1) % numRotations) * latheVertexList.length + 2;

			// Add top triangle
			triangleList[currentTriangle++] = new Triangle3d(color, TOP, a, c);

			// Diagram of bottom triangle being added:
			//
			//  b   d
			//  \---|---/
			//   \**|  /
			//    \*| /
			//     \|/
			//      BOTTOM

			// Calculate vertex indices
			int b = a + latheVertexList.length - 1;
			int d = c + latheVertexList.length - 1;			
			
			// Add bottom triangle			
			triangleList[currentTriangle++] = new Triangle3d(color, BOTTOM, d, b);
		}

		// For each lathe vertex
		for (int j = 0; j < latheVertexList.length - 1; ++j)
		{
			// For each rotation
			for (int i = 0; i < numRotations; ++i)
			{
				// Diagram of side triangles being added:
				//
				//     a        c
				//     /\------/\------/
				//    /**\****/  \    /
				//   /****\**/    \  /
				//  /------\/------\/
				// b       d

				// Calculate vertex indices
				int a = ((i + 0) % numRotations) * latheVertexList.length + j + 2;
				int b = a + 1;
				int c = ((i + 1) % numRotations) * latheVertexList.length + j + 2;
				int d = c + 1;

				// Add triangles
				triangleList[currentTriangle++] = new Triangle3d(color, a, b, d);
				triangleList[currentTriangle++] = new Triangle3d(color, a, d, c);
			}
		}
	}

	public Model3d transform(Transform3d transform3d)
	{
		// For each vertex
		for (int i = 0; i < vertexList.length; ++i)
		{
			// Translate vertex
			vertexList[i] = transform3d.transform(vertexList[i]);
		}

		return this;
	}

	/** Draws light-sourced model to graphics context with Lambert shading and back-face culling.
	 *  @param g              graphics context.
	 *  @param imageSize      size of image being drawn. Used to scale 3d to 2d projection formula.
	 *  @param lightDirection normalised light-source direction vector.
	 *  @param glint          value between 0.0 (none) and 1.0 (maximum) indicating amount of light glint.
	 *  @param fov            size of Field of Vision (FOV) angle.
	 */
	void draw(Graphics g, Dimension imageSize, Vector3d lightDirection, float glint, float fov)
	{
		// Calculate projection plane dimensions
		float planeWidth  = (float)imageSize.width  / (2 * (float)Math.tan(fov * 0.5f));
		float planeHeight = (float)imageSize.height / (2 * (float)Math.tan(fov * 0.5f));

		// Screen positions of vertices
		int[] screenX = new int[vertexList.length];
		int[] screenY = new int[vertexList.length];
		
		// For each vertex
		for (int i = 0; i < vertexList.length; ++i)
		{
			// Project vertex 3d position to 2d screen position
			screenX[i] = (int)((vertexList[i].x * planeWidth ) / vertexList[i].z + imageSize.width  / 2);
			screenY[i] = (int)((vertexList[i].y * planeHeight) / vertexList[i].z + imageSize.height / 2);
		}

		// Current triangle vertex screen positions
		int[] xList = new int[3];
		int[] yList = new int[3];

		// For each triangle
		for (int i = 0; i < triangleList.length; ++i)
		{
			// Get vertex screen positions
			xList[0] = screenX[triangleList[i].vertex1];
			yList[0] = screenY[triangleList[i].vertex1];
			xList[1] = screenX[triangleList[i].vertex2];
			yList[1] = screenY[triangleList[i].vertex2];
			xList[2] = screenX[triangleList[i].vertex3];
			yList[2] = screenY[triangleList[i].vertex3];

			// If triangle is visible (screen co-rds in clockwise order)
			if ((xList[2] - xList[0]) * (yList[2] - yList[1]) - (yList[2] - yList[0]) * (xList[2] - xList[1]) > 0)
			{
				// Get triangle vertices
				Vector3d vertex1 = vertexList[triangleList[i].vertex1];
				Vector3d vertex2 = vertexList[triangleList[i].vertex2];
				Vector3d vertex3 = vertexList[triangleList[i].vertex3];

				// Calculate triangle normal
				Vector3d normal = Vector3d.getNormal(vertex1, vertex2, vertex3);
				// Calculate triangle brightness based on angle between normal and light-source direction
				float brightness = normal.dot(lightDirection);

				// Make brightness 0 if negative
				brightness = Math.max(brightness, 0);

				// Store triangle color
				Color color = triangleList[i].color;
				// Store amount absent colors will be amplified to provide glint
				float amplify = brightness * brightness * glint;

				// Calculate light-sourced triangle color
				int red   = (int)(color.getRed  () * brightness + (255 - color.getRed  ()) * amplify);
				int green = (int)(color.getGreen() * brightness + (255 - color.getGreen()) * amplify);
				int blue  = (int)(color.getBlue () * brightness + (255 - color.getBlue ()) * amplify);

				// Draw triangle
				g.setColor(new Color(red, green, blue));
				g.fillPolygon(xList, yList, 3);
			}
		}
	}
}